import pytest

from en16931.bank_info import BankInfo
from en16931.entity import Entity
from en16931.postal_address import PostalAddress


class TestEntity:

    def test_initialization(self):
        e = Entity(name="Acme Inc.", tax_scheme="VAT",
                   tax_scheme_id="ES34626691F", country="ES",
                   party_legal_entity_id="ES34626691F",
                   registration_name="Acme INc.", mail="acme@acme.io",
                   endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                   address="easy street", postalzone="08080",
                   city="Barcelona")
        assert e.is_valid()
        assert e.tax_scheme == "VAT"
        assert e.tax_scheme_id == "ES34626691F"
        assert e.endpoint == "ES76281415Y"
        assert e.registration_name == "Acme INc."

    def test_unsuported_tax_scheme(self):
        e = Entity()
        with pytest.raises(ValueError):
            e.tax_scheme = "ASF"

    def test_not_apostal_address(self):
        e = Entity()
        with pytest.raises(TypeError):
            e.postal_address = "some address"

    def test_invalid_entity(self):
        e = Entity(name="Asdf Inc.")
        assert not e.is_valid()

    def test_valid_entity(self):
        e = Entity()
        e.name = "Acme Inc."
        e.tax_scheme = "VAT"
        e.tax_scheme_id = "ES34626691F"
        e.country = "ES"
        e.party_legal_entity_id = "ES34626691F"
        e.registration_name = "Acme INc."
        e.endpoint = "ES76281415Y"
        e.endpoint_scheme = "ES:VAT"
        p = PostalAddress(address="easy street", city_name="Barcelona",
                          postal_zone="08080", country="ES")
        e.postal_address = p
        assert e.is_valid()
        assert e.tax_scheme == "VAT"
        assert e.tax_scheme_id == "ES34626691F"
        assert e.endpoint == "ES76281415Y"
        assert e.registration_name == "Acme INc."

    def test_bank_info(self):
        seller = Entity(name="Acme Inc.", tax_scheme="VAT",
                        tax_scheme_id="ES34626691F", country="ES",
                        party_legal_entity_id="ES34626691F",
                        registration_name="Acme INc.", mail="acme@acme.io",
                        endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                        address="easy street", postalzone="08080",
                        city="Barcelona")
        bank_info_seller = BankInfo(iban="ES661234563156", bic="AAAABBCCDDD")
        seller.bank_info = bank_info_seller
        assert seller.is_valid()
        assert seller.bank_info.iban == "ES661234563156"
        assert seller.bank_info.bic == "AAAABBCCDDD"

    def test_invalid_bank_info(self):
        e = Entity()
        with pytest.raises(ValueError):
            e.bank_info = BankInfo()

    def test_wrong_bank_info(self):
        e = Entity()
        with pytest.raises(TypeError):
            e.bank_info = "asdf"



import pytest

from en16931.invoice_line import InvoiceLine


class TestInvoiceLine:

    def test_initialization(self):
        il = InvoiceLine(quantity=11, unit_code="EA", price=2,
                         item_name='test', currency="EUR",
                         tax_percent=0.21, tax_category="S")
        assert il.is_valid()
        assert il.currency == "EUR"

    def test_invalid_currency(self):
        il = InvoiceLine()
        with pytest.raises(KeyError):
            il.currency = "blah"

    def test_no_taxes(self):
        il = InvoiceLine()
        assert il.tax is None

    def test_creation(self):
        il = InvoiceLine()
        il.quantity = 11
        il.price = 2
        il.item_name = 'test'
        il.tax_percent = 0.21
        il.tax_category = "S"
        assert il.is_valid()

    def test_line_extension_amount(self):
        il = InvoiceLine(quantity=11, unit_code="EA", price=2,
                         item_name='test', currency="EUR",
                         tax_percent=0.21, tax_category="S")
        assert str(il.line_extension_amount) == '22.00'

    def test_invalid_unit_code(self):
        il = InvoiceLine()
        with pytest.raises(ValueError):
            il.unit_code = "ASF"

    def test_invalid_price(self):
        il = InvoiceLine()
        with pytest.raises(ValueError):
            il.price = "dasdas"

    def test_invalid_line_extension_amount(self):
        il = InvoiceLine()
        with pytest.raises(ValueError):
            il.line_extension_amount = "expensive"

    def test_invalid_quantity(self):
        il = InvoiceLine()
        with pytest.raises(ValueError):
            il.quantity = "dasdas"



import pytest

from en16931.invoice import Invoice


class TestImportInvoices:
    
    def test_imports_invoice1_xml(self, xml_path):
        invoice = Invoice.from_xml(xml_path)
        assert invoice.invoice_id == '1'
        assert invoice.issue_date.strftime("%Y-%m-%d") == "2018-06-11"
        assert invoice.due_date.strftime("%Y-%m-%d") == "2018-09-11"
        # Seller and buyer
        assert invoice.seller_party.is_valid()
        assert invoice.buyer_party.is_valid()
        # totals
        assert str(invoice.line_extension_amount) == '87.00'
        assert str(invoice.tax_exclusive_amount) == '87.00'
        assert str(invoice.tax_inclusive_amount) == '103.62'
        assert str(invoice.payable_amount) == '103.62'
        # Check imported lines by computing the values of
        # totals from lines and comparing them to those from
        # the XML.
        assert invoice.line_extension_amount == invoice.subtotal()
        assert invoice.tax_exclusive_amount == invoice.subtotal()
        assert invoice.tax_inclusive_amount == invoice.total()
        assert invoice.payable_amount == invoice.total()
        # Check that we stored the original xml
        with open(xml_path, 'rb') as fh:
            assert invoice.to_xml().encode('utf8') == fh.read()

    def test_imports_invoice2_xml(self, xml_path_invoice2):
        invoice = Invoice.from_xml(xml_path_invoice2)
        assert invoice.invoice_id == '1'
        assert invoice.issue_date.strftime("%Y-%m-%d") == "2018-06-11"
        assert invoice.due_date.strftime("%Y-%m-%d") == "2018-09-11"
        # Seller and buyer
        assert invoice.seller_party.is_valid()
        assert invoice.buyer_party.is_valid()
        # totals
        assert str(invoice.line_extension_amount) == '60.00'
        assert str(invoice.tax_exclusive_amount) == '50.00'
        assert str(invoice.tax_inclusive_amount) == '60.50'
        assert str(invoice.payable_amount) == '60.50'
        # discount and charge
        assert str(invoice.charge_amount) == '10.00'
        assert str(invoice.discount_amount) == '20.00'
        # Check imported lines by computing the values of
        # totals from lines and charges and discounts and
        # comparing them to those from the XML.
        assert invoice.tax_exclusive_amount == invoice.subtotal()
        assert invoice.tax_inclusive_amount == invoice.total()
        assert invoice.payable_amount == invoice.total()
        # Check that we stored the original xml
        with open(xml_path_invoice2, 'rb') as fh:
            assert invoice.to_xml().encode('utf8') == fh.read()



"""
Class for representing an Invoice.
"""
from datetime import datetime
import lxml.etree

from jinja2 import Environment, PackageLoader
from money.currency import Currency

from en16931.entity import Entity
from en16931.money import MyMoney
from en16931.utils import parse_date
from en16931.utils import parse_money
from en16931.xpaths import get_from_xpath
from en16931.xpaths import get_entity
from en16931.xpaths import get_invoice_lines
from en16931.xpaths import get_discount
from en16931.xpaths import get_charge


templates = Environment(loader=PackageLoader('en16931', 'templates'))

VALID_PAYMENT_CODES = {
    '10': 'cash',
    '49': 'debit',
    '31': 'transfer',
    '26': 'cheque',
    '23': 'cheque_b',
    '48': 'credit',
    'ZZZ': 'awarding, reposition, special',
}


class Invoice:
    """EN16931 Invoice class.

    This is the main entry point of this library. You can build, step
    by step, an Invoice and then serialize to XML.

    It uses the class :class:`Entity` to represent seller and buyer
    parties, and the class :class:`InvoiceLine` to represent
    invoice lines.

    >>> from en16931 import Invoice
    >>> invoice = Invoice(invoice_id="2018-01", currency="EUR")
    >>> seller = Entity(name="Acme Inc.", tax_scheme="VAT",
    ...                 tax_scheme_id="ES34626691F", country="ES",
    ...                 party_legal_entity_id="ES34626691F",
    ...                 registration_name="Acme INc.", mail="acme@acme.io",
    ...                 endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
    ...                 address="easy street", postalzone="08080",
    ...                 city="Barcelona")
    >>> buyer = Entity(name="Corp Inc.", tax_scheme="VAT",
    ...                tax_scheme_id="ES76281415Y", country="ES",
    ...                party_legal_entity_id="ES76281415Y",
    ...                registration_name="Corp INc.", mail="corp@corp.io",
    ...                endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
    ...                address="busy street", postalzone="08080",
    ...                city="Barcelona")
    >>> invoice.buyer_party = buyer
    >>> invoice.seller_party = seller
    >>> invoice.due_date = "2018-09-11"
    >>> invoice.issue_date = "2018-06-11"
    >>> # lines
    >>> il1 = InvoiceLine(quantity=11, unit_code="EA", price=2,
    ...                   item_name='test 1', currency="EUR",
    ...                   tax_percent=0.21, tax_category="S")
    >>> il2 = InvoiceLine(quantity=2, unit_code="EA", price=25,
    ...                   item_name='test 2', currency="EUR",
    ...                   tax_percent=0.21, tax_category="S")
    >>> il3 = InvoiceLine(quantity=5, unit_code="EA", price=3,
    ...                   item_name='test 3', currency="EUR",
    ...                   tax_percent=0.1, tax_category="S")
    >>> invoice.add_lines_from([il1, il2, il3])
 
    And serialize it to XML:

    >>> # As a string
    >>> xml = invoice.to_xml()
    >>> # Or save it directly to a file
    >>> invoice.save('example_invoice.xml')

    """

    def __init__(self, invoice_id=None, currency="EUR", from_xml=False):
        """Initialize an Invoice.

        This is the main class and entry point for creating an Invoice.


        Parameters
        ----------
        invoice_id: string (optional, default '1')
            Arbitrary string to identify the invoice.

        currency: string (optional, default 'EUR')
            An ISO 4217 currency code.

        from_xml: bool (optional, default False)
            A flag to mark if the object is the result of importing
            an XML invoice.


        Raises
        ------
        KeyError: If the currency code is not a valid ISO 4217 code.


        Examples
        --------

        By default the currency of the invoice is EUR and its id is 1:

        >>> i = Invoice()
        >>> i.invoice_id
        1
        >>> i.currency
        EUR

        You can also specify an arbitrary id and a valid ISO 4217
        currency code.

        >>> i = Invoice(invoice_id="0001-2018", currency="USD")
        >>> i.invoice_id
        0001-2018
        >>> i.currency
        USD

        """
        self.invoice_id = invoice_id or '1'
        self.currency = currency
        self.ubl_version_id = "2.1"
        self.customization_id = "urn:cen.eu:en16931:2017#compliant#urn:fdc:peppol.eu:2017:poacc:billing:3.0"
        self.profile_id = "urn:fdc:peppol.eu:2017:poacc:billing:01:1.0"
        self.invoice_type_code = 380
        self._issue_date = None
        self._due_date = None
        self._seller_party = None
        self._buyer_party = None
        self._templates = templates.get_template('invoice.xml')
        self._imported_from_xml = from_xml
        self._line_extension_amount = None
        self._tax_exclusive_amount = None
        self._tax_inclusive_amount = None
        self._payable_amount = None
        self._charge_amount = None
        self._charge_percent = None
        self._discount_amount = None
        self._discount_percent = None
        self._original_xml = None
        self._payment_means_code = None
        self.lines = []

    @classmethod
    def from_xml(cls, xml_path):
        """Import a XML invoice in EN16931 format.


        Parameters
        ----------
        xml_path: path
            A path to the XML file.


        Raises
        ------
        FileNotFoundError: if the file does not exist.


        Examples
        --------

        >>> i = Invoice.from_xml('path/to/invoice.xml')

        """
        with open(xml_path, "rb") as f:
            xml = f.read()
        root = lxml.etree.fromstring(xml)
        invoice_id = get_from_xpath(root, "invoice_id")
        currency = get_from_xpath(root, "currency")
        invoice = cls(invoice_id=invoice_id, currency=currency, from_xml=True)
        invoice._original_xml = xml.decode('utf8')
        invoice.issue_date = get_from_xpath(root, "invoice_issue_date")
        invoice.due_date = get_from_xpath(root, "invoice_due_date")
        # seller and buyer
        invoice.seller_party = get_entity(root, kind='seller')
        invoice.buyer_party = get_entity(root, kind='buyer')
        # totals
        invoice.line_extension_amount = get_from_xpath(root, "invoice_line_extension_amount")
        invoice.tax_exclusive_amount = get_from_xpath(root, "tax_exclusive_amount")
        invoice.tax_inclusive_amount = get_from_xpath(root, "tax_inclusive_amount")
        invoice.payable_amount = get_from_xpath(root, "payable_amount")
        # lines
        invoice.add_lines_from(get_invoice_lines(root))
        # discount and charge
        discount = get_discount(root)
        if discount is not None:
            invoice.discount_amount = discount
        charge = get_charge(root)
        if charge is not None:
            invoice.charge_amount = charge
        return invoice

    @property
    def currency(self):
        """Property: String representation of the ISO 4217 currency code.

        Parameters
        ----------
        currency_str: string
            String representation of the ISO 4217 currency code.

        Raises
        ------
        KeyError: If the currency code is not a valid ISO 4217 code.

        """
        return self._currency.name

    @currency.setter
    def currency(self, currency_str):
        try:
            self._currency = Currency[currency_str]
        except KeyError:
            raise KeyError('Currency {} not suported'.format(currency_str))

    def to_xml(self):
        """Serialize the invoice object to XML.

        Generate a valid PEPPOL BIS 3 XML document using the UBL 2.1
        syntax.
        """
        if self._original_xml is not None:
            return self._original_xml
        return self._templates.render(invoice=self)

    def save(self, path=None):
        """Save the XML representation of the invoice.


        Parameters
        ----------
        path: a path (optional, default None)
            If the path is None it a file named 'invoice_id.xml'
            will be created in the current working directory.

        """
        if path is None:
            path = 'invoice_{}.xml'.format(self.invoice_id)
        xml = self._original_xml if self._original_xml is not None else self.to_xml()
        with open(path, 'w') as f:
            f.write(xml)

    @property
    def payment_means_code(self):
        """Property: The payment means code.

        It has to be one of:
            * '10': 'cash'
            * '49': 'debit'
            * '31': 'transfer'
            * '26': 'cheque'
            * '23': 'cheque_b'
            * '48': 'credit'
            * 'ZZZ': 'awarding, reposition, special'

        Parameters
        ----------
        code: string
            A valid payment means code.

        Raises
        ------
        ValueError
            If the code is not valid.

        """
        return self._payment_means_code

    @payment_means_code.setter
    def payment_means_code(self, code):
        """Sets the payment means code
        """
        if code not in VALID_PAYMENT_CODES:
            raise ValueError('Invalid code {}, see documentation'.format(code))
        self._payment_means_code = code

    @property
    def issue_date(self):
        """Property: The issue date of the invoice.

        Parameters
        ----------
        date: datetime or string
            If the input is a string, it should be in one of the
            following formats: "%Y-%m-%d", "%Y%m%d", "%d-%m-%Y",
            "%Y/%m/%d", "%d/%m/%Y".

        Raises
        ------
        ValueError:
            if the input string cannot be converted to a
            datetime object.

        Examples
        --------

        >>> from datetime import datetime
        >>> i = Invoice()
    
        Supported date formats are:

        >>> i.issue_date = datetime(2018, 6, 21)
        >>> i.issue_date
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.issue_date = "2018-06-21"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.issue_date = "20180621"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.issue_date = "21-6-2018"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.issue_date = "2018/06/21"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.issue_date = "21/6/2018"
        datetime.datetime(2018, 6, 21, 0, 0)

        Incorrect date formats will raise a ValueError:

        >>> i.issue_date = "today"
        Traceback (most recent call last):
        [...]
        ValueError: See documentation for string date formats supported

        """
        return self._issue_date

    @issue_date.setter
    def issue_date(self, date):
        """Set the issue date of the invoice.
        """
        if not date:
            return
        elif isinstance(date, datetime):
            self._issue_date = date
        elif isinstance(date, str):
            self._issue_date = parse_date(date)
        else:
            raise ValueError("Unrecognized date")

    @property
    def due_date(self):
        """Property: Due date of the invoice.

        Parameters
        ----------
        date: datetime or string
            If the input is a string, it should be in one of the
            following formats: "%Y-%m-%d", "%Y%m%d", "%d-%m-%Y",
            "%Y/%m/%d", "%d/%m/%Y".

        Raises
        ------
        ValueError
            if the input string cannot be converted to a
            datetime object.
        
        Examples
        --------

        >>> from datetime import datetime
        >>> i = Invoice()
    
        Supported date formats are:

        >>> i.due_date = datetime(2018, 6, 21)
        >>> i.due_date
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.due_date = "2018-06-21"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.due_date = "20180621"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.due_date = "21-6-2018"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.due_date = "2018/06/21"
        datetime.datetime(2018, 6, 21, 0, 0)
        >>> i.due_date = "21/6/2018"
        datetime.datetime(2018, 6, 21, 0, 0)

        Incorrect date formats will raise a ValueError:

        >>> i.due_date = "today"
        Traceback (most recent call last):
        [...]
        ValueError: See documentation for string date formats supported

        """
        return self._due_date

    @due_date.setter
    def due_date(self, date):
        """Set the due date of the invoice.
        """
        if not date:
            return
        elif isinstance(date, datetime):
            self._due_date = date
        elif isinstance(date, str):
            self._due_date = parse_date(date)
        else:
            raise ValueError("Unrecognized date")

    @property
    def seller_party(self):
        """Property: The Entity with the role of AccountingSupplierParty.

        See the :class:`Entity` class for details

        Parameters
        ----------
        party: Entity object.
            The Entity object that plays the role of
            AccountingSupplierParty.

        Raises
        ------
        ValueError:
            if the Entity is not valid.

        TypeError:
            if the input is not an Entity or Entity subclass.

        """
        return self._seller_party

    @seller_party.setter
    def seller_party(self, party):
        """Set the Entity with the role of AccountingSupplierParty.
        """
        if isinstance(party, Entity):
            if party.is_valid():
                self._seller_party = party
            else:
                raise ValueError("Invalid Entity")
        else:
            msg = "Expected an Entity object but got a {}"
            raise TypeError(msg.format(type(party)))

    @property
    def buyer_party(self):
        """Property: The Entity with the role of AccountingCustomerParty.

        See the :class:`Entity` class for details

        Parameters
        ----------
        party: Entity object.
            The Entity object that plays the role of
            AccountingCustomerParty.

        Raises
        ------
        ValueError
            if the Entity is not valid.

        TypeError
            if the input is not an Entity or Entity subclass.

        """
        return self._buyer_party

    @buyer_party.setter
    def buyer_party(self, party):
        """Set the Entity with the role of AccountingCustomerParty.
        """
        if isinstance(party, Entity):
            if party.is_valid():
                self._buyer_party = party
            else:
                raise ValueError("Invalid Entity")
        else:
            msg = "Expected an Entity object but got a {}"
            raise TypeError(msg.format(type(party)))

    @property
    def charge_amount(self):
        """Property: The ChargeTotalAmount of the Invoice.

        Parameters
        ----------
        value: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library.

        Raises
        ------
        decimal.InvalidOperation: If the input cannot be converted
            to a Decimal.

        """
        if self._charge_amount is not None:
            return self._charge_amount
        else:
            return MyMoney('0', self._currency)

    @charge_amount.setter
    def charge_amount(self, value):
        """Sets the ChargeTotalAmount of the invoice.
        """
        self._charge_amount = parse_money(value, self._currency)
        self._charge_percent = round(self._charge_amount / self.gross_subtotal(), 2)

    @property
    def charge_percent(self):
        """The percentage that the charge represents.

        The MultiplierFactorNumeric of the charge in PEPPOL BIS 3 terms.

        Parameters
        ----------
        value: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library.

        """
        return self._charge_percent

    @charge_percent.setter
    def charge_percent(self, value):
        """Sets the Charge percent of the invoice.
        """
        if value > 1 or value < -1:
            self._charge_percent = round(value / 100, 2)
        else:
            self._charge_percent = round(value, 2)
        amount = self._charge_percent * self.gross_subtotal()
        self._charge_amount = parse_money(amount, self._currency)

    @property
    def charge_base_amount(self):
        """The base amount of the charge.

        The BaseAmount of the charge in PEPPOL BIS 3 terms.
        """
        if self._charge_amount is not None and self._charge_percent is not None:
            return self._charge_amount / self._charge_percent

    @property
    def discount_amount(self):
        """Property: The AllowanceTotalAmount of the Invoice.

        Parameters
        ----------
        value: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library.

        Raises
        ------
        decimal.InvalidOperation: If the input cannot be converted
            to a Decimal.

        """
        if self._discount_amount is not None:
            return self._discount_amount
        else:
            return MyMoney('0', self._currency)

    @discount_amount.setter
    def discount_amount(self, value):
        """Sets the AllowanceTotalAmount of the invoice.
        """
        self._discount_amount = parse_money(value, self._currency)
        self._discount_percent = round(self._discount_amount / self.gross_subtotal(), 2)

    @property
    def discount_percent(self):
        """The percentage that the discount represents.

        The MultiplierFactorNumeric of the discount in PEPPOL BIS 3 terms.
        """
        return self._discount_percent

    @discount_percent.setter
    def discount_percent(self, value):
        """Sets the discount percent of the invoice.
        """
        if value > 1 or value < -1:
            self._discount_percent = round(value / 100, 2)
        else:
            self._discount_percent = round(value, 2)
        amount = self._discount_percent * self.gross_subtotal()
        self._discount_amount = parse_money(amount, self._currency)

    @property
    def discount_base_amount(self):
        """The base amount of the discount.

        The BaseAmount of the discount in PEPPOL BIS 3 terms.
        """
        if self._discount_amount is not None and self._discount_percent is not None:
            return self._discount_amount / self._discount_percent

    def add_line(self, line):
        """Adds an InvoiceLine to the Invoice.

        Parameters
        ----------
        line: InvoiceLine object.

        """
        self.lines.append(line)

    def add_lines_from(self, container):
        """Adds InvoiceLine instances from a container.

        Parameters
        ----------
        container: container
            An iterable container of InvoiceLine objects.

        """
        self.lines.extend(container)

    @property
    def unique_taxes(self):
        """Set of unique taxes in the Invoice.
        """
        return {line.tax for line in self.lines}

    def lines_with_taxes(self, tax_type=None):
        """Generator of InvoiceLines

        Parameters
        ----------
        tax_type: Tax object (default None).
            If a Tax object is provided, only generate lines with that
            Tax. If this parameter is None, generate all lines.
        """
        for line in self.lines:
            if line.has_tax(tax_type):
                yield line

    def tax_amount(self, tax_type=None):
        """Computes the tax amount of the Invoice.

        Parameters
        ----------
        tax_type: Tax object (default None).
            If a Tax object is provided, the tax amount corresponding
            to the porvided Tax. If None the total tax amount.

        """
        if tax_type is None:
            taxes = self.unique_taxes
        else:
            taxes = {tax_type}
        result = (self.taxable_base(tax_type=tax) * tax.percent for tax in taxes)
        return sum(result, MyMoney('0', self._currency))

    def taxable_base(self, tax_type=None):
        """Computes the taxable base of the Invoice

        Parameters
        ----------
        tax_type: Tax object (default None).
            If a Tax object is provided, the taxable base corresponding
            to the porvided Tax. If None the total taxable base.

        """
        return self.gross_subtotal(tax_type=tax_type) - self.discount_amount + self.charge_amount

    def gross_subtotal(self, tax_type=None):
        """Sum of gross amount of each invoice line."""
        amounts = (line.line_extension_amount for line in
                   self.lines_with_taxes(tax_type=tax_type))
        return sum(amounts, MyMoney('0', self._currency))

    def subtotal(self, tax_type=None):
        """Gross amount before taxes.

            TotalGrossAmount - AllowanceTotalAmount + ChargeTotalAmount
        """
        gross_subtotal = self.gross_subtotal(tax_type=tax_type)
        return gross_subtotal - self.discount_amount + self.charge_amount

    def total(self):
        """Computes the TaxInclusiveAmount of the Invoice
        """
        return self.subtotal() + self.tax_amount()

    # Properties so we can return what was in the XML instead of computing it
    # in case we read the invoice.
    @property
    def line_extension_amount(self):
        """The total LineExtensionAmount of the invoice.

        It's only computed as the :meth:`gross_subtotal` if the Invoice
        was not imported from an XML file. In that case, its value is
        the one reported on the XML.
        """
        if self._line_extension_amount is not None:
            return self._line_extension_amount
        return self.gross_subtotal()

    @line_extension_amount.setter
    def line_extension_amount(self, value):
        """Sets the LineExtensionAmount of the invoice.

        Only used when importing from an XML file.
        """
        self._line_extension_amount = parse_money(value, self._currency)

    @property
    def tax_exclusive_amount(self):
        """The total TaxExclusiveAmount of the invoice.

        It's only computed as the :meth:`gross_subtotal` if the Invoice
        was not imported from an XML file. In that case, its value is
        the one reported on the XML.
        """
        if self._tax_exclusive_amount is not None:
            return self._tax_exclusive_amount
        return self.subtotal()

    @tax_exclusive_amount.setter
    def tax_exclusive_amount(self, value):
        """Sets the TaxExclusiveAmount of the invoice.

        Only used when importing from an XML file.
        """
        self._tax_exclusive_amount = parse_money(value, self._currency)

    @property
    def tax_inclusive_amount(self):
        """The total TaxInclusiveAmount of the invoice.

        It's only computed as the :meth:`total` if the Invoice
        was not imported from an XML file. In that case, its value is
        the one reported on the XML.
        """
        if self._tax_inclusive_amount is not None:
            return self._tax_inclusive_amount
        return self.total()

    @tax_inclusive_amount.setter
    def tax_inclusive_amount(self, value):
        """Sets the TaxInclusiveAmount of the invoice.

        Only used when importing from an XML file.
        """
        self._tax_inclusive_amount = parse_money(value, self._currency)

    @property
    def payable_amount(self):
        """The total PayableAmount of the invoice.

        It's only computed as the :meth:`total` if the Invoice
        was not imported from an XML file. In that case, its value is
        the one reported on the XML.
        """
        if self._payable_amount is not None:
            return self._payable_amount
        # TODO PrepaidAmount
        prepaid_amount = MyMoney('0', self._currency)
        return self.total() - prepaid_amount

    @payable_amount.setter
    def payable_amount(self, value):
        """Sets the PayableAmount of the invoice.

        Only used when importing from an XML file.
        """
        self._payable_amount = parse_money(value, self._currency)



"""
Module to interact with b2brouter.net
"""

import requests

IMPORT_URL = "https://app.b2brouter.net/projects/{}/invoices/xml.json"
HALTR_URL = "https://app.b2brouter.net"
IMPORT_URL_TEST = "http://localhost:3001/projects/{}/invoices/xml.json"
HALTR_URL_TEST = "http://localhost:3001"


def post_to_b2brouter(invoice, api_key, project_id, test=False):
    """Posts an Invoice to b2brouter.net

    Parameters
    ----------
    api_key: string.
        The authentification API key for b2brouter.net

    project_id: string.
        The project ID to which submit the invoice in b2brouter.net

    """
    payload = invoice.to_xml().encode('utf8')
    import_url = IMPORT_URL.format(project_id) if not test else \
                 IMPORT_URL_TEST.format(project_id)
    headers = {
        'content-type': 'application/octet-stream',
        'X-Redmine-API-Key': api_key,
    }
    response = requests.post(import_url, data=payload, headers=headers)
    if response.status_code == 201:
        return response.json()['invoice']['id']
    elif response.status_code == 401:
        raise RuntimeError("Unauthorized: invalid API_KEY")
    else:
        raise RuntimeError("\n".join(response.json()['errors']))



"""
Class for representing a Postal Address
"""
class PostalAddress:
    """PostalAddress class

    It represents a postal address of an :class:`Entity`.
    """

    def __init__(self, address=None, city_name=None,
                 postal_zone=None, country=None):
        """Initializes a PostalAddress.

        Parameters
        ----------
        address: string.
            An address.

        city_name: string.
            The name of a city.

        postal_zone: string.
            A valid postal zone.

        country: string.
            A valid two letter country code.

        """
        self.address = address
        self.city_name = city_name
        self.postal_zone = postal_zone
        self.country = country

    @property
    def country(self):
        """The country of the address
        """
        return self._country

    @country.setter
    def country(self, country):
        """Sets the country of the address.
        """
        # TODO validate country in ISO3166-1:Alpha2 format
        self._country = country



"""
Miscelanious util functions
"""
from datetime import datetime
from decimal import InvalidOperation

from en16931.money import MyMoney

def parse_money(amount, currency):
    if amount is None:
        return None
    elif isinstance(amount, MyMoney):
        return amount
    try:
        return MyMoney(amount, currency)
    except InvalidOperation:
        raise ValueError('Could not convert {} to money'.format(amount))


def parse_float(flt):
    if flt is None:
        return None
    try:
        return float(flt)
    except ValueError:
        raise ValueError('Could not convert {} to float'.format(flt))


def parse_date(date):
    formats = [
        "%Y-%m-%d",
        "%Y%m%d",
        "%d-%m-%Y",
        "%Y/%m/%d",
        "%d/%m/%Y",
    ]
    for fmt in formats:
        try:
            return datetime.strptime(date, fmt)
        except ValueError:
            continue
    else:
        raise ValueError("See documentation for string date formats supported")



import pytest
import responses

from en16931 import validex


api_key='6df3de0197ee4f9a8b296c51dcc54a5c'
user_id=5


def test_validex_success(invoice1):
    validex_success_json = {'report': {'documentHash': '933098de',
        'documentTypes': [{'criteria': [{'expression': "(contains(namespace-uri(), 'urn:oasis:names:specification:ubl:schema:xsd') and ends-with(namespace-uri(), '2'))",
            'expressionType': 'xpath'}],
            'matches': True,
            'name': 'UBL 2',
            'path': 'doc/xml/ubl2',
            'subStart': True},
            {'criteria': [{'expression': "namespace-uri() = 'urn:oasis:names:specification:ubl:schema:xsd:Invoice-2'",
                'expressionType': 'xpath'}],
                'matches': True,
                'name': 'Invoice',
                'path': 'doc/xml/ubl2/inv',
                'subStart': False}],
            'duration': 0.47534894943237,
            'filename': '1531216744.xml',
            'id': 'ljc7yi',
            'result': 'success',
            'success': True,
            'timestamp': '2018-07-10T10:06:09+00:00',
            'userId': 5,
            'validationSteps': [{'description': 'XML Structural Validation',
                'documentType': 'doc/xml',
                'errors': [],
                'name': 'xml-structure',
                'success': True,
                'validationType': 'xml'},
                {'description': 'UBL 2.1 Invoice',
                    'documentType': 'doc/xml/ubl2/inv',
                    'errors': None,
                    'name': 'ubl-2_1-invoice-schema',
                    'success': True,
                    'validationType': 'schema'}]}} 
    responses.add(responses.POST, 'https://api2.validex.net/api/validate',
                  json=validex_success_json, status=200)
    assert validex.is_valid_at_validex(invoice1, api_key, user_id)


def test_validex_fail(invoice1):
    validex_success_json = {'report': {'documentHash': '933098de',
        'documentTypes': [{'criteria': [{'expression': "(contains(namespace-uri(), 'urn:oasis:names:specification:ubl:schema:xsd') and ends-with(namespace-uri(), '2'))",
            'expressionType': 'xpath'}],
            'matches': True,
            'name': 'UBL 2',
            'path': 'doc/xml/ubl2',
            'subStart': True},
            {'criteria': [{'expression': "namespace-uri() = 'urn:oasis:names:specification:ubl:schema:xsd:Invoice-2'",
                'expressionType': 'xpath'}],
                'matches': True,
                'name': 'Invoice',
                'path': 'doc/xml/ubl2/inv',
                'subStart': False}],
            'duration': 0.47534894943237,
            'filename': '1531216744.xml',
            'id': 'ljc7yi',
            'result': 'fatal',
            'success': True,
            'timestamp': '2018-07-10T10:06:09+00:00',
            'userId': 5,
            'validationSteps': [{'description': 'XML Structural Validation',
                'documentType': 'doc/xml',
                'errors': [],
                'name': 'xml-structure',
                'success': True,
                'validationType': 'xml'},
                {'description': 'UBL 2.1 Invoice',
                    'documentType': 'doc/xml/ubl2/inv',
                    'errors': None,
                    'name': 'ubl-2_1-invoice-schema',
                    'success': True,
                    'validationType': 'schema'}]}} 
    responses.add(responses.POST, 'https://api2.validex.net/api/validate',
                  json=validex_success_json, status=200)
    assert validex.is_valid_at_validex(invoice1, api_key, user_id)



"""
Module to interact with open.validex.net

You need to create an user at validex.net to be able to
use its API.

"""

import base64
import requests
import time

URL = "https://api2.validex.net/api/validate"


def is_valid_at_validex(invoice, api_key, user_id):
    """Validates an Invoice at open.validex.net

    You need to create an user at validex.net to be able to
    use its API.

    Parameters
    ----------
    api_key: string.
        The authentification API key for validex.net

    user_id: string.
        The user ID of validex.net

    Notes
    -----
    Warnings are not reported.

    """
    payload = {
        'userId': user_id,
        'filename': "{}.xml".format(int(time.time())),
        'fileContents64': base64.b64encode(invoice.to_xml().encode('utf8')),
    }
    headers = {
        'content_type': 'json',
        'accept': 'json',
        'Authorization': "apikey={}".format(api_key),
    }
    response = requests.post(URL, json=payload, headers=headers, verify=False)
    if response.status_code == 200:
        result = response.json()
        if result.get('report') and result['report'].get('result') and\
                result['report']['result'] != 'fatal':
            # Warnings are not reported
            return True
        elif result.get('report') and result['report'].get('id'):
            raise ValueError(_get_validation_errors(result['report']['id']))
        else:
            raise RuntimeError(": ".join(response.json()['error'].values()))
    elif response.status_code == 401:
        raise RuntimeError("Unauthorized: invalid API_KEY")
    else:
        raise RuntimeError(": ".join(response.json()['error'].values()))


def _get_validation_errors(report_id):
    headers = {
        'accept': 'json',
        'Authorization': "apikey={}".format(api_key),
    }
    response = requests.get(
        "{}/{}".format(URL.replace('validate', 'report'), report_id),
        headers=headers,
        verify=False,
    )
    if response.status_code == 200:
        resp = response.json()
        errors = []
        validation_steps = []
        if resp['report'].get('validationSteps'):
            validation_steps.extend(resp['report']['validationSteps'])
        for step in validation_steps:
            if step.get('success') or not step.get('errors'):
                continue
            for error in step['errors']:
                if error.get('message'):
                    errors.add("{}: {}".format(step['description'], error['message']))
                elif error.get('text'):
                    errors.add("{}: {}".format(step['description'], error['text']))
                else:
                    errors.add(str(error))
        return ". ".join(errors)
    elif response.status_code == 401:
        raise RuntimeError("Unauthorized: invalid API_KEY")
    else:
        raise RuntimeError(": ".join(response.json()['error'].values()))



"""
Class for representing an invoice line.
"""
from money.money import Money
from money.currency import Currency

from en16931.tax import Tax
from en16931.utils import parse_float
from en16931.utils import parse_money


UNIT_CODES = {
    'EA': 'units',
    'HUR': 'hours',
    'KGM': 'kilograms',
    'LTR': 'litters',
    'DAY': 'days',
    'CS': 'boxes',
}


class InvoiceLine:
    """EN16931 InvoiceLine class.

    Each :class:`Invoice` has to have at least one invoice
    line in which the quantity and the price of the items is
    reflected.

    You can initialize an InvoiceLine instance with all its
    attributes:

    >>> il = InvoiceLine(quantity=11, unit_code="EA", price=2,
    ...                  item_name='test', currency="EUR",
    ...                  tax_percent=0.21, tax_category="S")

    Or cou can do it step by step:

    >>> il = InvoiceLine()
    >>> il.quantity = 11
    >>> il.price = 2
    >>> il.item_name = 'test'
    >>> il.tax_percent = 0.21
    >>> il.tax_category = "S"

    An InvoiceLine is only valid if it has quantity, price and
    tax defined:

    >>> il.is_valid()
    True
    >>> new_line = InvoiceLine()
    >>> new_line.is_valid()
    False

    """

    def __init__(self, quantity=None, unit_code="EA", price=None,
                 item_name=None, currency="EUR", tax_percent=None,
                 line_extension_amount=None, tax_category=None,
                 tax_name=None):
        """Initialize an Invoice Line.

        Parameters
        ----------
        quantity: float or integer.
            The number of items of the line.

        unit_code: string (optional).
            A unit code defining the nature of the quantities of the
            items of the line. It must be one of: 'EA': 'units',
            'HUR': 'hours', 'KGM': 'kilograms', 'LTR': 'litters',
            'DAY': 'days', 'CS': 'boxes'.

        price: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library.

        item_name: string (optional).
            Arbitrary name to define the item of the line.

        currency: string.
            String representation of the ISO 4217 currency code.

        tax_percent: float.
            The percentage of the Tax applied to the line.
            Can be 0.

        tax_category: string.
            A string representing the category of the Tax.
            It must be one of 'AE', 'L', 'M', 'E', 'S', 'Z',
            'G', 'O', or 'K'.

        tax_name: string.
            Arbitrary name to identify the Tax.

        line_extension_amount: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library. Computed unless the invoice
            is imported from an XML file.


        Notes
        -----
        An InvoiceLine is considered valid if and only if it has quantity,
        price and tax.

        """
        self.currency = currency
        self.item_name = item_name
        self.quantity = quantity
        self.unit_code = unit_code
        self.price = price
        self.line_extension_amount = line_extension_amount
        self.tax_percent = tax_percent
        self.tax_category = tax_category
        self.tax_name = tax_name

    @property
    def currency(self):
        """Property: String representation of the ISO 4217 currency code.

        Parameters
        ----------
        currency_str: string
            String representation of the ISO 4217 currency code.

        Raises
        ------
        KeyError: If the currency code is not a valid ISO 4217 code.

        """
        return self._currency.name

    @currency.setter
    def currency(self, currency_str):
        """Sets the currency of the Invoice.
        """
        try:
            self._currency = Currency[currency_str]
        except KeyError:
            raise KeyError('Currency {} not suported'.format(currency_str))

    @property
    def tax(self):
        """Returns a Tax object representing the taxes applied to the line.
        """
        if self.tax_percent and self.tax_category:
            return Tax(self.tax_percent, self.tax_category, self.tax_name or "")
        else:
            return None

    @property
    def item_name(self):
        """The arbitrary name of the item of the line.

        Parameters
        ----------
        item_name: string (optional).
            Arbitrary name to define the item of the line.

        """
        return self._item_name

    @item_name.setter
    def item_name(self, name):
        """Sets the arbitrary name of the item of the line.
        """
        self._item_name = name

    @property
    def quantity(self):
        """Property: Quantity of items of the line.

        Parameters
        ---------
        quantity: float or integer.
            The number of items of the line.

        """
        return self._quantity

    @quantity.setter
    def quantity(self, quantity):
        """Sets the quantity of items of the line.
        """
        try:
            self._quantity = parse_float(quantity)
        except ValueError:
            raise ValueError("Unrecognized quantity {}".format(quantity))

    @property
    def price(self):
        """Property: The price of one item.

        Parameters
        ---------
        price: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library.

        """
        return self._price

    @price.setter
    def price(self, price):
        """Sets the price of one item.
        """
        if price is None:
            return
        try:
            self._price = parse_money(price, self._currency)
            self._line_extension_amount = self._price * self._quantity
        except ValueError:
            raise ValueError("Unrecognized price {}".format(price))

    @property
    def line_extension_amount(self):
        """Property: The LineExtensionAmount

        Parameters
        ---------
        line_extension_amount: string, integer, float
            The input must be a valid input for the Decimal class
            the Python Standard Library. Computed unless the invoice
            is imported from an XML file.


        """
        if self._line_extension_amount is not None:
            return self._line_extension_amount
        else:
            return self._price * self._quantity

    @line_extension_amount.setter
    def line_extension_amount(self, price):
        """Sets the LineExtensionAmount.
        """
        if price is None:
            return
        try:
            self._line_extension_amount = parse_money(price, self._currency)
        except ValueError:
            raise ValueError("Unrecognized line_extension_amount {}".format(price))

    @property
    def unit_code(self):
        """property: The unit code defining the nature of the quantities.

        Parameters
        ----------
        unit_code: string.
            A unit code defining the nature of the quantities of the
            items of the line. It must be one of: 'EA': 'units',
            'HUR': 'hours', 'KGM': 'kilograms', 'LTR': 'litters',
            'DAY': 'days', 'CS': 'boxes'.

        """
        return self._unit_code

    @unit_code.setter
    def unit_code(self, code):
        """Sets the unit code defining the nature of the quantities.
        """
        if code not in UNIT_CODES:
            raise ValueError("Unsupported unit code {}".format(code))
        self._unit_code = code

    def is_valid(self):
        """Returns True if the line is valid.
        """
        has_quantity = self._quantity is not None
        has_price = self._price is not None
        has_tax = self.tax is not None
        return has_quantity and has_price and has_tax

    def has_tax(self, tax):
        """Returns True if the line has this tax.

        Parameters
        ----------
        tax: Tax Object.
            
        """
        if tax is None:
            return True
        return self.tax == tax

    def __repr__(self):
        if not self.is_valid():
            return "{}: empty".format(self.__class__)
        return "{}: {} {} x {} {}".format(self.__class__, self.quantity,
                                          self.item_name, self.price,
                                          self.currency)



"""
Class for representing a Tax category.
"""
from en16931.utils import parse_float


CATEGORIES = {'AE', 'L', 'M', 'E', 'S', 'Z', 'G', 'O', 'K'}


class Tax:
    """Tax class.

    It representas a tax to apply globally or to a concrete
    invoice line.

    Only categories of taxes enabled by the EN16931 standard are
    supported. See the documentation of :meth:`category` property
    for more details.

    You can create Tax objects directly:

    >>> t = Tax(0.21, "S", "IVA")

    Or specify the relevant attributes when building
    :class:`InvoiceLines` or :class:`Invoice`

    """

    def __init__(self, percent, category, name, comment=""):
        """Initialize a Tax object.

        Parameters
        ----------
        category: string.
            A string representing the category of the Tax.
            It must be one of 'AE', 'L', 'M', 'E', 'S', 'Z',
            'G', 'O', or 'K'.

        percent: float.
            The percentage of the Tax. Can be 0.

        name: string.
            Arbitrary name to identify the Tax.

        comment: string.
            A comment on the tax.

        Notes
        -----
        A tax is compared to other Tax objects by equality of their
        percentage, category, and name.

        """
        self.category = category
        self.name = name
        self.comment = comment
        pct = parse_float(percent)
        if pct > 1 or pct < -1:
            self.percent = pct / 100
        else:
            self.percent = pct

    @property
    def category(self):
        """Property: The category of the Tax.

        Parameters
        ----------
        category: string.
            A string representing the category of the Tax.
            It must be one of 'AE', 'L', 'M', 'E', 'S', 'Z',
            'G', 'O', or 'K'.

        Raises
        ------
        ValueError: if the category is not valid.

        """
        return self._category

    @category.setter
    def category(self, category):
        """Sets the category of the Tax.
        """
        if category not in CATEGORIES:
            msg = "Category {} not valid. Use one of {}"
            raise ValueError(msg.format(category, CATEGORIES))
        self._category = category

    @property
    def code(self):
        """An identification code of the tax.
        """
        return "{}_{}".format(self.percent, self.category)

    def __eq__(self, other):
        """
        A tax is compared to other Tax objects by equality of their
        percentage, category, and name.
        """
        if other is None:
            return False
        return (self.percent == other.percent and
                self.category == self.category and
                self.name == self.name)

    def __hash__(self):
        return hash(self.code)

    def __repr__(self):
        return "Tax {}: {} {} {}".format(self.category, self.percent,
                                         self.name, self.comment)



"""
Bank information Class
"""

class BankInfo:
    """BankInfo class.

    Stores relevant banking information of :class:`Entity`
    to specify the needed information when the payment means
    of the invoice involves a Bank.

    You can initialize a BankInfo instance with all needed
    attributes:
    
    >>> b = BankInfo(account="1234567321", bic="AAAABBCCDDD",
    ...              mandate_reference_identifier="123")

    Or build it, step by step:

    >>> b = BankInfo()
    >>> b.account = "1234567321"
    >>> b.bic = "AAAABBCCDDD"
    >>> b.mandate_reference_identifier = "123"

    For the bank information to be complete it has to contain the
    IBAN number, or the bank account and the BIC. The mandate
    reference identifier is used in the context of debit payments.

    >>> b.is_valid()
    True

    """

    def __init__(self, account=None, iban=None, bic=None,
                 mandate_reference_identifier=None):
        """Initialize the Bank Information for an Entity.

        Parameters
        ----------
        account: string (optional)
            The bank account number.

        bic: string (optional)
            The Bank Identification Code of the account.

        iban: string (optional)
            The International Bank Account Number (IBAN)

        mandate_reference_identifier: string (optional)
            The Mandate Reference Identifier

        Notes
        -----
        For the bank information to be complete it has to contain the
        IBAN number, or the bank account and the BIC. The mandate
        reference identifier is used in the context of debit payments.

        """
        self.account = account
        self.iban = iban
        self.bic = bic
        self.mandate_reference_identifier = mandate_reference_identifier

    @property
    def account(self):
        """Property: Bank account number

        Parameters
        ----------
        account: string
            The bank account number.

        Notes
        -----
        No validation of the bank account number is performed.

        """
        return self._account

    @account.setter
    def account(self, account):
        """Sets the bank account number
        """
        self._account = account

    @property
    def iban(self):
        """Property: The International Bank Account Number (IBAN)

        Parameters
        ----------
        iban: string
            The International Bank Account Number (IBAN)

        Notes
        -----
        No validation of the IBAN is performed.

        """
        return self._iban

    @iban.setter
    def iban(self, iban):
        """Sets the IBAN
        """
        self._iban = iban

    @property
    def bic(self):
        """Property: The Bank Identification Code

        Also known as SWIFT code.

        Parameters
        ----------
        bic: string
            The Bank Identification Code of the account.

        Raises
        ------
        ValueError
            If the BIC code is not valid.

        Notes
        -----
        A BIC code has either 8 or 11 characters without
        dashes and spaces.

        """
        return self._bic

    @bic.setter
    def bic(self, bic):
        """Sets the Bank Identification Code
        """
        if bic is None or len(bic) == 8 or len(bic) == 11:
            self._bic = bic
        else:
            raise ValueError("A BIC code has either 8 or 11 charcters")

    @property
    def mandate_reference_identifier(self):
        """Property: The Mandate Reference Identifier

        The Mandate Reference Identifier is necessary for debit
        payments.

        Parameters
        ----------
        mandate_reference_identifier: string
            The Mandate Reference Identifier

        """
        return self._mandate_reference_identifier

    @mandate_reference_identifier.setter
    def mandate_reference_identifier(self, mandate_reference_identifier):
        """Sets the Mandate Reference Identifier
        """
        self._mandate_reference_identifier = mandate_reference_identifier

    def is_valid(self):
        """Returns True if the information is complete.

        A valid bank information for an :class:`Entity` must have a bank
        account and a BIC or an IBAN.
        """
        has_account = self._account is not None and self._bic is not None
        has_iban = self._iban is not None
        return has_account or has_iban

    def has_mandate_reference(self):
        """Returns True if it has a mandate reference identifier.

        This is necessary for debit payments.
        """
        return self._mandate_reference_identifier is not None



import os
import pytest
from datetime import datetime
from decimal import Decimal

from en16931.entity import Entity
from en16931.invoice import Invoice
from en16931.invoice_line import InvoiceLine
from en16931.tax import Tax


class TestInvoiceAttributes:

    def test_default_id_number(self):
        i = Invoice()
        assert i.invoice_id == '1'

    def test_id_number(self):
        i = Invoice(invoice_id="1-2018")
        assert i.invoice_id == "1-2018"

    def test_set_issue_date(self):
        dts = {
            "2018-09-11": "%Y-%m-%d",
            "20180911": "%Y%m%d",
            "21-06-2018": "%d-%m-%Y",
            "2018/12/12": "%Y/%m/%d",
            "12/12/2018": "%d/%m/%Y",
        }
        i = Invoice()
        for date, fmt in dts.items():
            i.issue_date = date
            assert i.issue_date.strftime(fmt) == date

    def test_set_dates_none(self):
        i = Invoice()
        i.issue_date = None
        assert i.issue_date is None
        i.due_date = None
        assert i.due_date is None

    def test_set_dates_datetime(self):
        i = Invoice()
        i.issue_date = datetime(2018, 7, 10)
        assert i.issue_date == datetime(2018, 7, 10)
        i.due_date = datetime(2018, 9, 10)
        assert i.due_date == datetime(2018, 9, 10)

    def test_wrong_issue_date(self):
        with pytest.raises(ValueError):
            i = Invoice()
            i.issue_date = "asdef"
        with pytest.raises(ValueError):
            i = Invoice()
            i.issue_date = 123

    def test_set_due_date(self):
        dts = {
            "2018-09-11": "%Y-%m-%d",
            "20180911": "%Y%m%d",
            "21-06-2018": "%d-%m-%Y",
            "2018/12/12": "%Y/%m/%d",
            "12/12/2018": "%d/%m/%Y",
        }
        i = Invoice()
        for date, fmt in dts.items():
            i.due_date = date
            assert i.due_date.strftime(fmt) == date

    def test_wrong_due_date(self):
        with pytest.raises(ValueError):
            i = Invoice()
            i.due_date = "asdef"
        with pytest.raises(ValueError):
            i = Invoice()
            i.due_date = 123

    def test_invalid_currency(self):
        with pytest.raises(KeyError):
            i = Invoice()
            i.currency = "asdef"

    def test_wrong_seller_entity(self):
        with pytest.raises(TypeError):
            i = Invoice()
            i.seller_party = 123
        with pytest.raises(ValueError):
            i = Invoice()
            i.seller_party = Entity()

    def test_wrong_buyer_entity(self):
        with pytest.raises(TypeError):
            i = Invoice()
            i.buyer_party = 123
        with pytest.raises(ValueError):
            i = Invoice()
            i.buyer_party = Entity()

    def test_invalid_payment_code(self):
        with pytest.raises(ValueError):
            i = Invoice()
            i.payment_means_code = "asdef"

    def test_payment_code(self):
            i = Invoice()
            i.payment_means_code = "10"
            assert i.payment_means_code == "10"


class TestInvoiceOperations:

    def test_unique_taxes(self, invoice1, tax1, tax2):
        assert len(invoice1.unique_taxes) == 2
        assert invoice1.unique_taxes == {tax1, tax2}

    def test_lines_with_taxes(self, invoice1, tax1, tax2):
        assert len(list(invoice1.lines_with_taxes(tax_type=tax1))) == 2
        assert len(list(invoice1.lines_with_taxes(tax_type=tax2))) == 1

    def test_tax_amount(self, invoice1, tax1, tax2):
        assert str(invoice1.tax_amount()) == '16.62'
        assert str(invoice1.tax_amount(tax_type=tax1)) == '15.12'
        assert str(invoice1.tax_amount(tax_type=tax2)) == '1.50'

    def test_taxable_base(self, invoice1, tax1, tax2):
        assert str(invoice1.taxable_base()) == '87.00'
        assert str(invoice1.taxable_base(tax_type=tax1)) == '72.00'
        assert str(invoice1.taxable_base(tax_type=tax2)) == '15.00'

    def test_gross_subtotal(self, invoice1):
        assert str(invoice1.gross_subtotal()) == '87.00'

    def test_subtotal(self, invoice1):
        assert str(invoice1.gross_subtotal()) == '87.00'

    def test_total(self, invoice1):
        assert str(invoice1.total()) == '103.62'

    def test_line_extension_amount(self, invoice1):
        assert str(invoice1.line_extension_amount) == '87.00'

    def test_tax_exclusive_amount(self, invoice1):
        assert str(invoice1.tax_exclusive_amount) == '87.00'

    def test_tax_inclusive_amount(self, invoice1):
        assert str(invoice1.tax_inclusive_amount) == '103.62'

    def test_payable_amount(self, invoice1):
        assert str(invoice1.payable_amount) == '103.62'

    def test_charge_amount(self, invoice1):
        assert str(invoice1.charge_amount) == '0.00'
        invoice1.charge_amount = 10
        assert str(invoice1.charge_amount) == '10.00'
        assert str(invoice1.subtotal()) == '97.00'

    def test_discount_amount(self, invoice1):
        assert str(invoice1.discount_amount) == '0.00'
        invoice1.discount_amount = 10
        assert str(invoice1.discount_amount) == '10.00'
        assert str(invoice1.subtotal()) == '77.00'

    def test_charge_percent(self, invoice1):
        invoice1.charge_percent = 10
        assert str(invoice1.charge_percent) == '0.1'
        assert str(invoice1.charge_amount) == '8.70'
        assert str(invoice1.subtotal()) == '95.70'

    def test_discount_percent(self, invoice1):
        invoice1.discount_percent = 10
        assert str(invoice1.discount_percent) == '0.1'
        assert str(invoice1.discount_amount) == '8.70'
        assert str(invoice1.subtotal()) == '78.30'



class TestInvoiceXMLGeneration:

    def test_generates_xml_invoice1(self, invoice1):
        out = invoice1.to_xml()
        assert len(out) > 0

    def test_generates_xml_invoice2(self, invoice2):
        out = invoice2.to_xml()
        assert len(out) > 0

    def test_generates_xml_invoice3(self, invoice3):
        out = invoice3.to_xml()
        assert len(out) > 0

    def test_writes_xml_file_invoice1(self, invoice1):
        path = '/tmp/invoice1.xml'
        invoice1.save(path)
        assert os.path.exists(path)

    def test_writes_xml_file_invoice2_with_discount_and_charge(self, invoice2):
        path = '/tmp/invoice2.xml'
        invoice2.save(path)
        assert os.path.exists(path)

    def test_writes_xml_file_invoice3_with_payment_means(self, invoice3):
        path = '/tmp/invoice3.xml'
        invoice3.save(path)
        assert os.path.exists(path)



import pytest
import os

from en16931.bank_info import BankInfo
from en16931.entity import Entity
from en16931.invoice import Invoice
from en16931.invoice_line import InvoiceLine
from en16931.tax import Tax


@pytest.fixture()
def xml_path():
    path, name = os.path.split(os.path.abspath(__file__))
    return os.path.join(path, "files", "invoice.xml")


@pytest.fixture()
def xml_path_invoice2():
    path, name = os.path.split(os.path.abspath(__file__))
    return os.path.join(path, "files", "invoice2.xml")


@pytest.fixture()
def invoice1():
    invoice = Invoice()
    seller = Entity(name="Acme Inc.", tax_scheme="VAT",
                    tax_scheme_id="ES34626691F", country="ES",
                    party_legal_entity_id="ES34626691F",
                    registration_name="Acme INc.", mail="acme@acme.io",
                    endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                    address="easy street", postalzone="08080",
                    city="Barcelona")
    buyer = Entity(name="Corp Inc.", tax_scheme="VAT",
                   tax_scheme_id="ES76281415Y", country="ES",
                   party_legal_entity_id="ES76281415Y",
                   registration_name="Corp INc.", mail="corp@corp.io",
                   endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                   address="busy street", postalzone="08080",
                   city="Barcelona")
    invoice.buyer_party = buyer
    invoice.seller_party = seller
    invoice.due_date = "2018-09-11"
    invoice.issue_date = "2018-06-11"
    # lines
    il1 = InvoiceLine(quantity=11, unit_code="EA", price=2,
                      item_name='test 1', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    il2 = InvoiceLine(quantity=2, unit_code="EA", price=25,
                      item_name='test 2', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    il3 = InvoiceLine(quantity=5, unit_code="EA", price=3,
                      item_name='test 3', currency="EUR",
                      tax_percent=0.1, tax_category="S")
    invoice.add_lines_from([il1, il2, il3])
    return invoice


@pytest.fixture()
def invoice2():
    invoice = Invoice()
    seller = Entity(name="Acme Inc.", tax_scheme="VAT",
                    tax_scheme_id="ES34626691F", country="ES",
                    party_legal_entity_id="ES34626691F",
                    registration_name="Acme INc.", mail="acme@acme.io",
                    endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                    address="easy street", postalzone="08080",
                    city="Barcelona")
    buyer = Entity(name="Corp Inc.", tax_scheme="VAT",
                   tax_scheme_id="ES76281415Y", country="ES",
                   party_legal_entity_id="ES76281415Y",
                   registration_name="Corp INc.", mail="corp@corp.io",
                   endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                   address="busy street", postalzone="08080",
                   city="Barcelona")
    invoice.buyer_party = buyer
    invoice.seller_party = seller
    invoice.due_date = "2018-09-11"
    invoice.issue_date = "2018-06-11"
    # lines
    il1 = InvoiceLine(quantity=5, unit_code="EA", price=2,
                      item_name='test 1', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    il2 = InvoiceLine(quantity=2, unit_code="EA", price=25,
                      item_name='test 2', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    invoice.add_lines_from([il1, il2])
    # discount and charge
    invoice.charge_amount = 10
    invoice.discount_amount = 20
    return invoice

@pytest.fixture()
def invoice3():
    invoice = Invoice()
    seller = Entity(name="Acme Inc.", tax_scheme="VAT",
                    tax_scheme_id="ES34626691F", country="ES",
                    party_legal_entity_id="ES34626691F",
                    registration_name="Acme INc.", mail="acme@acme.io",
                    endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                    address="easy street", postalzone="08080",
                    city="Barcelona")
    bank_info_seller = BankInfo(iban="ES661234563156", bic="AAAABBCCDDD")
    seller.bank_info = bank_info_seller
    buyer = Entity(name="Corp Inc.", tax_scheme="VAT",
                   tax_scheme_id="ES76281415Y", country="ES",
                   party_legal_entity_id="ES76281415Y",
                   registration_name="Corp INc.", mail="corp@corp.io",
                   endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
                   address="busy street", postalzone="08080",
                   city="Barcelona")
    bank_info_buyer = BankInfo(iban="ES661234567321", bic="AAAABBCCDDD",
                               mandate_reference_identifier="123")
    buyer.bank_info = bank_info_buyer
    invoice.buyer_party = buyer
    invoice.seller_party = seller
    invoice.due_date = "2018-09-11"
    invoice.issue_date = "2018-06-11"
    invoice.payment_means_code = "31"
    # lines
    il1 = InvoiceLine(quantity=11, unit_code="EA", price=2,
                      item_name='test 1', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    il2 = InvoiceLine(quantity=2, unit_code="EA", price=25,
                      item_name='test 2', currency="EUR",
                      tax_percent=0.21, tax_category="S")
    il3 = InvoiceLine(quantity=5, unit_code="EA", price=3,
                      item_name='test 3', currency="EUR",
                      tax_percent=0.1, tax_category="S")
    invoice.add_lines_from([il1, il2, il3])
    return invoice


@pytest.fixture()
def tax1():
    return Tax(0.21, "S", None)


@pytest.fixture()
def tax2():
    return Tax(0.1, "S", None)



"""
Class representing an Entity.
"""

from en16931.bank_info import BankInfo
from en16931.postal_address import PostalAddress


class Entity:
    """Entity class.

    It represents a party involved in the trade described by the
    invoice, such as seller or buyer party.

    You can initialize an Entity with most of its attributes:

    >>> e = Entity(name="Acme Inc.", tax_scheme="VAT",
    ...            tax_scheme_id="ES34626691F", country="ES",
    ...            party_legal_entity_id="ES34626691F",
    ...            registration_name="Acme INc.", mail="acme@acme.io",
    ...            endpoint="ES76281415Y", endpoint_scheme="ES:VAT",
    ...            address="easy street", postalzone="08080",
    ...            city="Barcelona")

    Or you can build it step by step:

    >>> e.name = "Acme Inc."
    >>> e.tax_scheme = "VAT"
    >>> e.tax_scheme_id = "ES34626691F"
    >>> e.country = "ES"
    >>> e.party_legal_entity_id = "ES34626691F"
    >>> e.registration_name = "Acme INc."
    >>> e.endpoint = "ES76281415Y"
    >>> e.endpoint_scheme = "ES:VAT"
    >>> p = PostalAddress(address="easy street", city_name="Barcelona",
    ...                   postal_zone="08080", country="ES")
    >>> e.postal_address = p

    You can assign a :class:`PostalAddress` to its :meth:`postal_address`
    property. You can also associate a :class:`BankInfo` instance to an
    Entity in order to store relevant banking information for debit and
    transfer payments:

    >>> bank_info = BankInfo(iban="ES661234563156", bic="AAAABBCCDDD")
    >>> e.bank_info = bank_info
    >>> e.bank_info.iban
    "ES661234563156"

    An entity is valid if it has a name, a country, valid ids,
    valid taxscheme and endpoint, and has an address.

    >>> e.is_Valid()
    True

    """

    def __init__(self, name=None, tax_scheme=None, tax_scheme_id=None, country=None,
                 party_legal_entity_id=None, registration_name=None, mail=None,
                 endpoint=None, endpoint_scheme=None, postalzone=None, city=None,
                 address=None):
        """Initialize an Entity.

        TODO formal definition of Entity.

        Parameters
        ----------
        name: string.
            The name of the Entity.

        tax_scheme: string.
            The tax scheme of the Entity.

        tax_scheme_id: string.
            The tax ID of the Entity.

        country: string.
            Two letter code for the country of the Entity.

        party_legal_entity_id: string.
            The party legal entity of the Entity.

        registration_name: string.
            The Registration name of the Entity.

        mail: string.
            The contact Email of the Entity.

        endpoint: string.
            A valid PEPPOL endpoint.

        endpoint_scheme: string.
            The scheme defining the endpoint.

        postalzone: string.
            The postalzone of the address of the Entity.

        city: string.
            The city of the address of the Entity.

        address: string.
            The address of the Entity.


        Notes
        -----
        An entity is valid if it has a name, a country, valid ids,
        valid taxscheme and endpoint, and has an address.

        """
        self.name = name
        self.tax_scheme = tax_scheme
        self.tax_scheme_id = tax_scheme_id
        self.endpoint = endpoint
        self.endpoint_scheme = endpoint_scheme
        self.country = country
        self.party_legal_entity_id = party_legal_entity_id
        self.registration_name = registration_name
        self.mail = mail
        self.postalzone = postalzone
        self.city = city
        self.address = address
        self._postal_address = None
        self._bank_info = None

    @property
    def bank_info(self):
        """Property: a :class:`BankInfo` instance

        Parameters
        ----------
        bank_info: BankInfo instance
            a valid BankInfo instance.

        Raises
        ------
        TypeError
            If the parameter is not a BankInfo instance
        ValueError
            If the BankInfo instance is not valid

        """
        return self._bank_info

    @bank_info.setter
    def bank_info(self, bank_info):
        """Sets the BankInfor instance
        """
        if isinstance(bank_info, BankInfo):
            if bank_info.is_valid():
                self._bank_info = bank_info
            else:
                raise ValueError("Invalid BankInfo instance")
        else:
            raise TypeError('Expected a BankInfo instance')

    @property
    def postal_address(self):
        """Property: The PostalAddress of the Entity.

        See the :class:`PostalAddress` class.

        Parameters
        ----------
        address: PostalAddress object.
            A PostalAddress instance.

        Raises
        ------
        TypeError: if the input is not a PostalAddreess
            or a subclass.

        """
        if self._postal_address is not None:
            return self._postal_address
        elif self.postalzone is None or self.country is None:
            return None
        else:
            return PostalAddress(address=self.address, city_name=self.city,
                                 postal_zone=self.postalzone, country=self.country)

    @postal_address.setter
    def postal_address(self, address):
        """Sets the PostalAddress of the Entity
        """
        if isinstance(address, PostalAddress):
            self._postal_address = address
        else:
            msg = "Expected a PostalAddress object but got a {}"
            raise TypeError(msg.format(type(address)))

    @property
    def name(self):
        """Property: The name of the Entity.

        Parameters
        ----------
        name: string.
            The name for the Entity.

        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of the Entity
        """
        self._name = name

    @property
    def tax_scheme(self):
        """Property: The tax scheme of the Entity.

        Parameters
        ----------
        scheme: string.
            The tax scheme used by the Entity.

        Raises
        ------
        ValueError: if the tax scheme is not valid.

        """
        return self._tax_scheme

    @tax_scheme.setter
    def tax_scheme(self, scheme):
        """Sets the tax scheme of the Entity.
        """
        supported_schemes = {None, "VAT"}
        if scheme not in supported_schemes:
            raise ValueError("Unsupported tax scheme %s" % scheme)
        self._tax_scheme = scheme

    @property
    def tax_scheme_id(self):
        """Property: The tax ID of the Entity.

        Parameters
        ----------
        tax_scheme_id: string.
            The tax ID of the Entity.

        """
        return self._tax_scheme_id

    @tax_scheme_id.setter
    def tax_scheme_id(self, tax_scheme_id):
        """Sets the tax ID of the Entity.
        """
        # TODO validate ID
        self._tax_scheme_id = tax_scheme_id

    @property
    def endpoint_scheme(self):
        """Property: The endpoint scheme of the Entity.

        Parameters
        ----------
        endpoint_scheme: string.
            The scheme defining the endpoint.

        """
        return self._endpoint_scheme

    @endpoint_scheme.setter
    def endpoint_scheme(self, scheme):
        """Sets the endpoint scheme of the Entity.
        """
        # TODO validate scheme
        self._endpoint_scheme = scheme

    @property
    def endpoint(self):
        """Property: The endpoint ID of the Entity.

        Parameters
        ----------
        endpoint: string.
            A valid PEPPOL endpoint.

        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        """Sets the endpoint ID of the Entity
        """
        # TODO validate endpoint
        self._endpoint = endpoint

    @property
    def country(self):
        """Property: The country of the entity.

        Parameters
        ----------
        country: string.
            Two letter code for the country of the Entity.

        """
        return self._country

    @country.setter
    def country(self, country):
        """Sets the country of the entity.
        """
        # TODO validate country
        self._country = country

    @property
    def party_legal_entity_id(self):
        """The party legal entity ID
        """
        return self._party_legal_entity_id

    @party_legal_entity_id.setter
    def party_legal_entity_id(self, party_legal_entity_id):
        """Sets the party legal entity ID.

        Parameters
        ----------
        party_legal_entity_id: string.
            Party legal entity ID

        """
        # TODO validate party_legal_entity_id
        self._party_legal_entity_id = party_legal_entity_id

    @property
    def registration_name(self):
        """The registration name of the Entity.
        """
        return self._registration_name

    @registration_name.setter
    def registration_name(self, registration_name):
        """Sets the registration name of the entity.

        Parameters
        ----------
        registration_name: string.
            The Registration name of the Entity.

        """
        # TODO validate registration_name
        self._registration_name = registration_name

    @property
    def mail(self):
        """The contact mail of the Entity.
        """
        return self._mail

    @mail.setter
    def mail(self, mail):
        """Sets the contact mail of the Entity.
        """
        # TODO validate mail
        self._mail = mail

    def is_valid(self):
        """Returns True if the Entity is valid.

        An entity is valid if it has a name, a country, valid ids,
        valid taxscheme and endpoint, and has an address.
        """
        has_name = self._name is not None
        has_country = self._country is not None
        has_ids = (self._party_legal_entity_id is not None) and (self._tax_scheme_id is not None)
        has_valid_taxscheme = self._tax_scheme is not None
        has_endpoint = self._endpoint is not None
        has_address = self.postal_address is not None
        return (has_name and has_country and has_ids and has_valid_taxscheme
                and has_endpoint and has_address)



"""
python-en16931
==============

Python package to read, write and manage the new EN16931 Invoice format:
The European Norm that describes a pan-European semantic model for the
invoices and their mapping to two standard syntaxes, UBL and CII.

"""
from en16931.bank_info import BankInfo
from en16931.entity import Entity
from en16931.invoice import Invoice
from en16931.invoice_line import InvoiceLine
from en16931.tax import Tax
from en16931.postal_address import PostalAddress
from en16931 import xpaths
from en16931 import b2brouter
from en16931 import utils
from en16931 import validex



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="en16931",
    version="0.1",
    author="Invinet Sistemes",
    author_email="jtorrents@ingent.net",
    description="A Python 3 package to parse, generate and manage the EN16931 Invoice format",
    long_description=long_description,
    url="https://github.com/invinet/python-en16931",
    packages=setuptools.find_packages(),
    classifiers=(
	"Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
)



import pytest
from collections import Hashable

from en16931.tax import Tax


class TestTaxes:

    def test_initialization(self):
        t = Tax(0.21, "S", "IVA")
        assert t

    def test_hashable(self):
        t = Tax(0.21, "S", "IVA")
        assert isinstance(t, Hashable)

    def test_percent_less_than_one(self):
        t = Tax(0.21, "S", "IVA")
        assert t.percent == 0.21

    def test_percent_more_than_one(self):
        t = Tax(21, "S", "IVA")
        assert t.percent == 0.21

    def test_percent_string(self):
        t = Tax("21", "S", "IVA")
        assert t.percent == 0.21

    def test_cmp_with_None(self):
        t = Tax("21", "S", "IVA")
        assert not (t == None)

    def test_value_error_bad_percent(self):
        with pytest.raises(ValueError):
            t = Tax("asdf", "S", "IVA")

    def test_value_error_bad_category(self):
        with pytest.raises(ValueError):
            t = Tax("21", "asd", "IVA")



"""
Manipulate and parse XML files
"""

import lxml.etree

from en16931.entity import Entity
from en16931.invoice_line import InvoiceLine

xpaths = None
namespaces = None


def get_discount(root, namespaces=None, xpaths=None):
    """Gets the discount of an Invoice
    """
    if namespaces is None:
        namespaces = get_namespaces()
    if xpaths is None:
        xpaths = get_xpaths()
    for line in root.findall('cac:AllowanceCharge', namespaces=namespaces):
        if get_from_xpath(line, "acharge_indicator") == "false":
            return get_from_xpath(line, "acharge_amount")


def get_charge(root, namespaces=None, xpaths=None):
    """Gets the charge of an Invoice
    """
    if namespaces is None:
        namespaces = get_namespaces()
    if xpaths is None:
        xpaths = get_xpaths()
    for line in root.findall('cac:AllowanceCharge', namespaces=namespaces):
        if get_from_xpath(line, "acharge_indicator") == "true":
            return get_from_xpath(line, "acharge_amount")


def get_invoice_lines(root, namespaces=None):
    """Generator of InvoiceLines of an Invoice
    """
    if namespaces is None:
        namespaces = get_namespaces()
    keys = {
        'quantity': 'line_quantity',
        'unit_code': 'line_unit',
        'line_extension_amount': 'line_extension_amount',
        'item_name': 'line_item_name',
        'price': 'line_price',
        'currency': 'line_currency',
        'tax_percent': 'line_tax_percent',
        'tax_category': 'line_tax_category',
    }
    lines = root.findall('cac:InvoiceLine', namespaces=namespaces)
    for line in lines:
        args = {key:get_from_xpath(line, value) for key, value in keys.items()}
        yield InvoiceLine(**args)


def get_entity(root, kind='seller'):
    """Gets an Entity of an Invoice
    """
    if kind == 'seller':
        keys = {
	    'party_legal_entity_id': 'seller_party_legal_entity_id',
	    'name': 'seller_name',
	    'registration_name': 'seller_registration_name',
	    'tax_scheme_id': 'seller_taxcode',
	    'tax_scheme': 'seller_tax_scheme',
	    'endpoint_scheme': 'seller_endpoint_scheme',
	    'endpoint': 'seller_endpoint',
	    'country': 'seller_countrycode',
	    'address': 'seller_address',
	    'city': 'seller_city',
	    'postalzone': 'seller_postalzone',
	    'mail': 'seller_email',
	}
    else:
        keys = {
	    'party_legal_entity_id': 'buyer_party_legal_entity_id',
	    'name': 'buyer_name',
	    'registration_name': 'buyer_registration_name',
	    'tax_scheme_id': 'buyer_taxcode',
	    'tax_scheme': 'buyer_tax_scheme',
	    'endpoint_scheme': 'buyer_endpoint_scheme',
	    'endpoint': 'buyer_endpoint_scheme',
	    'country': 'buyer_countrycode',
	    'address': 'buyer_address',
	    'city': 'buyer_city',
	    'postalzone': 'buyer_postalzone',
	    'mail': 'buyer_email',
	}
    args = {key:get_from_xpath(root, value) for key, value in keys.items()}
    return Entity(**args)


def get_from_xpath(root, tag, xpaths=None, namespaces=None):
    """Gets the content of an XPATH in an XML file.
    """
    if xpaths is None:
        xpaths = get_xpaths()
    if namespaces is None:
        namespaces = get_namespaces()
    xpath = xpaths.get(tag)
    if xpath is None:
        raise KeyError("Unknown key {} for xpath".format(tag))
    result = root.xpath(xpath, namespaces=namespaces)
    if not result:
        return None
    text_result = [e.text if isinstance(e, lxml.etree._Element) else e for e in result]
    return ", ".join(text_result)


def get_namespaces():
    """Get a dictionari with all namespaces.
    """
    global namespaces
    if namespaces is None:
        #print("loading namespace")
        namespaces = en16931_namespaces()
    return namespaces


def en16931_namespaces():
    """Namespaces for the en16931 invoice format
    """
    namespaces = {
        "xmlns":"urn:oasis:names:specification:ubl:schema:xsd:Invoice-2",
        "cac":"urn:oasis:names:specification:ubl:schema:xsd:CommonAggregateComponents-2",
        "cbc":"urn:oasis:names:specification:ubl:schema:xsd:CommonBasicComponents-2",
    }
    return namespaces


def get_xpaths():
    """Get a dictionary with all xpaths
    """
    global xpaths
    if xpaths is None:
        #print("loading xpaths")
        xpaths = en16931_xpaths()
    return xpaths


def en16931_xpaths():
    """Xpaths for the en16931 invoice format.
    """
    xpaths = {}
    xpaths["invoice_id"] = "/xmlns:Invoice/cbc:ID"
    xpaths["invoice_date"] = "/xmlns:Invoice/cbc:IssueDate"
    xpaths["invoice_issue_date"] = "/xmlns:Invoice/cbc:IssueDate"
    xpaths["invoice_due_date"] = "/xmlns:Invoice/cbc:DueDate"
    xpaths["extra_info"] = "/xmlns:Invoice/cbc:Note"
    xpaths["tax_point_date"] = "/xmlns:Invoice/cbc:TaxPointDate"
    xpaths["invoice_total"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:TaxInclusiveAmount"
    xpaths["invoice_import"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:TaxExclusiveAmount"
    xpaths["tax_inclusive_amount"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:TaxInclusiveAmount"
    xpaths["tax_exclusive_amount"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:TaxExclusiveAmount"
    xpaths["payable_amount"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:PayableAmount"
    xpaths["invoice_line_extension_amount"] = "/xmlns:Invoice/cac:LegalMonetaryTotal/cbc:LineExtensionAmount"
    xpaths["payment_method"] = "/xmlns:Invoice/cac:PaymentMeans/cbc:PaymentMeansCode"
    xpaths["payment_means_due_date"] = "/xmlns:Invoice/cac:PaymentMeans/cbc:PaymentDueDate"
    xpaths["seller_taxcode"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyTaxScheme/cbc:CompanyID"
    xpaths["seller_tax_scheme"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyTaxScheme/cac:TaxScheme/cbc:ID"
    xpaths["seller_party_legal_entity_id"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyLegalEntity/cbc:CompanyID"
    xpaths["seller_registration_name"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyLegalEntity/cbc:RegistrationName"
    xpaths["seller_endpoint"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cbc:EndpointID"
    xpaths["seller_endpoint_scheme"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cbc:EndpointID/@schemeID"
    xpaths["seller_name"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyName/cbc:Name"
    xpaths["seller_address"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:StreetName"
    xpaths["seller_address2"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:AdditionalStreetName"
    xpaths["seller_province"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:CountrySubentity"
    xpaths["seller_postalzone"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:PostalZone"
    xpaths["seller_countrycode"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cac:Country/cbc:IdentificationCode"
    xpaths["seller_website"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cbc:WebsiteURI"
    xpaths["seller_email"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:Contact/cbc:ElectronicMail"
    xpaths["seller_city"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:CityName"
    xpaths["seller_cp_city"] = [ "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:PostalZone",
        "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:CityName" ]
    xpaths["seller_contact"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:Contact/cbc:Name"
    xpaths["seller_endpoint_id"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cbc:EndpointID"
    xpaths["seller_company_id"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PartyLegalEntity/cbc:CompanyID"
    xpaths["buyer_contactname"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:Contact/cbc:Name"
    xpaths["buyer_phone"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:Contact/cbc:Telephone"
    xpaths["buyer_taxcode"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyTaxScheme/cbc:CompanyID"
    xpaths["buyer_taxcode_id"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyIdentification/cbc:ID"
    xpaths["buyer_tax_scheme"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyTaxScheme/cac:TaxScheme/cbc:ID"
    xpaths["buyer_endpoint_id"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cbc:EndpointID"
    xpaths["buyer_endpoint_scheme"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cbc:EndpointID/@schemeID"
    xpaths["buyer_party_legal_entity_id"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyLegalEntity/cbc:CompanyID"
    xpaths["buyer_registration_name"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyLegalEntity/cbc:RegistrationName"
    xpaths["buyer_name"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PartyName/cbc:Name"
    xpaths["buyer_address"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:StreetName"
    xpaths["buyer_address2"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:AdditionalStreetName"
    xpaths["buyer_province"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:CountrySubentity"
    xpaths["buyer_postalzone"] = "/xmlns:Invoice/cac:AccountingSupplierParty/cac:Party/cac:PostalAddress/cbc:PostalZone"
    xpaths["buyer_countrycode"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cac:Country/cbc:IdentificationCode"
    xpaths["buyer_website"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cbc:WebsiteURI"
    xpaths["buyer_reference"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:Contact/cbc:ID"
    xpaths["buyer_email"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:Contact/cbc:ElectronicMail"
    xpaths["buyer_city"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:CityName"
    xpaths["buyer_cp_city"] = [ "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:PostalZone",
    "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:PostalAddress/cbc:CityName" ]
    xpaths["buyer_contact"] = "/xmlns:Invoice/cac:AccountingCustomerParty/cac:Party/cac:Contact/cbc:Name"
    xpaths["currency"] = "/xmlns:Invoice/cbc:DocumentCurrencyCode"

    xpaths["global_taxes"] = ["/xmlns:Invoice/cac:TaxTotal/cac:TaxSubtotal",
    "/xmlns:Invoice/cac:WithholdingTaxTotal/cac:TaxSubtotal"]
    # relative to global_taxes
    xpaths["gtax_category"] = "cac:TaxCategory/cbc:ID"
    xpaths["gtax_percent"] = "cac:TaxCategory/cbc:Percent"
    xpaths["gtax_name"] = "cac:TaxCategory/cac:TaxScheme/cbc:ID"

    xpaths["invoice_lines"] = "//cac:InvoiceLine"
    # relative to invoice_lines
    xpaths["i_transaction_ref"] = "IssuerTransactionReference" # todo
    xpaths["r_contract_reference"] = "ReceiverContractReference" # todo
    xpaths["line_quantity"] = "cbc:InvoicedQuantity"
    xpaths["line_description"] = ["cac:Item/cbc:Name", "cac:Item/cbc:Description"]
    xpaths["line_price"] = "cac:Price/cbc:PriceAmount"
    xpaths["line_currency"] = "cbc:LineExtensionAmount/@currencyID"
    xpaths["line_extension_amount"] = "cbc:LineExtensionAmount"
    xpaths["line_unit"] = "cbc:InvoicedQuantity/@unitCode"
    xpaths["line_taxes"] = ["cac:Item/cac:ClassifiedTaxCategory"]
    xpaths["line_tax_percent"] = "cac:Item/cac:ClassifiedTaxCategory/cbc:Percent"
    xpaths["line_tax_category"] = "cac:Item/cac:ClassifiedTaxCategory/cbc:ID"
    xpaths["line_tax_scheme"] = "cac:Item/cac:ClassifiedTaxCategory/cac:TaxScheme/cbc:ID"
    xpaths["line_item_name"] = "cac:Item/cbc:Name"
    xpaths["line_notes"] = ["cbc:Note","cac:Item/cac:SellersItemIdentification/cbc:ID"]
    xpaths["line_code"] = "cac:Item/cac:SellersItemIdentification/cbc:ID"
    xpaths["line_code2"] = "cac:Item/cac:StandardItemIdentification/cbc:ID"
    xpaths["line_code2_scheme"] = "cac:Item/cac:StandardItemIdentification/cbc:ID/@schemeID"
    xpaths["line_discounts"] = "cac:AllowanceCharge/cbc:ChargeIndicator[text()='false']/../../cac:AllowanceCharge"
    xpaths["line_charges"] = "cac:AllowanceCharge/cbc:ChargeIndicator[text()='true']/../../cac:AllowanceCharge"
    xpaths["sequence_number"] = "SequenceNumber" # todo
    xpaths["tax_event_code"] = "SpecialTaxableEvent/SpecialTaxableEventCode" # todo
    xpaths["tax_event_reason"] = "SpecialTaxableEvent/SpecialTaxableEventReason" # todo
    xpaths["delivery_notes"] = "DeliveryNotesReferences/DeliveryNote" # todo
    # relative to invoice_lines/delivery_notes_references/delivery_note
    xpaths["delivery_note_num"] = "DeliveryNoteNumber" # todo

    xpaths["ponumber"] = "/xmlns:Invoice/cac:OrderReference/cbc:ID"
    xpaths["contract_number"] = "/xmlns:Invoice/cac:ContractDocumentReference/cbc:ID"
    # relative to invoice_lines/discounts
    #xpaths["line_discount_percent"] = "cbc:MultiplierFactorNumeric"
    xpaths["line_discount_text"] = "cbc:AllowanceChargeReason"
    xpaths["line_discount_amount"] = "cbc:Amount"
    # relative to invoice_lines/charges
    xpaths["line_charge"] = "cbc:Amount"
    xpaths["line_charge_reason"] = "cbc:AllowanceChargeReason"
    # relative to invoice_lines/taxes
    xpaths["tax_name"] = "cac:TaxScheme/cbc:ID"
    xpaths["tax_category"] = "cbc:ID"
    xpaths["tax_percent"] = "cbc:Percent"

    xpaths["attachments"] = "//cac:AdditionalDocumentReference"
    # relative to attachment
    xpaths["attach_format"] = "cac:Attachment/cbc:EmbeddedDocumentBinaryObject/@mimeCode"
    xpaths["attach_encoding"] = "cac:Attachment/cbc:EmbeddedDocumentBinaryObject/@encodingCode"
    xpaths["attach_description"] = "cbc:DocumentType"
    xpaths["attach_data"] = "cac:Attachment/cbc:EmbeddedDocumentBinaryObject"
    xpaths["attach_filename"] = "cac:Attachment/cbc:EmbeddedDocumentBinaryObject/@filename"

    xpaths["delivery"] = "//cac:Delivery"
    # relative to delivery
    xpaths["delivery_date"] = "cbc:ActualDeliveryDate"
    xpaths["delivery_location_type"] = "cac:DeliveryLocation/cbc:ID/@schemeID"
    xpaths["delivery_location_id"] = "(cac:DeliveryLocation//cbc:ID)[1]"
    xpaths["delivery_address"] = "cac:DeliveryLocation/cac:Address/cbc:StreetName"
    xpaths["delivery_address2"] = "cac:DeliveryLocation/cac:Address/cbc:AdditionalStreetName"
    xpaths["delivery_city"] = "cac:DeliveryLocation/cac:Address/cbc:CityName"
    xpaths["delivery_postalcode"] = "cac:DeliveryLocation/cac:Address/cbc:PostalZone"
    xpaths["delivery_province"] = "cac:DeliveryLocation/cac:Address/cbc:CountrySubentity"
    xpaths["delivery_country"] = "cac:DeliveryLocation/cac:Address/cac:Country/cbc:IdentificationCode"

    xpaths["acharge"] = "/xmlns:Invoice/cac:AllowanceCharge"
    # relative to charge, may be a charge or discount
    xpaths["acharge_indicator"] = "cbc:ChargeIndicator"
    xpaths["acharge_amount"] = "cbc:Amount"
    xpaths["acharge_reason"] = "cbc:AllowanceChargeReason"
    xpaths["acharge_tax_c"] = "cac:TaxCategory/cbc:ID"
    xpaths["acharge_tax_p"] = "cac:TaxCategory/cbc:Percent"
    xpaths["acharge_tax_n"] = "cac:TaxCategory/cac:TaxScheme/cbc:ID"

    return xpaths



from decimal import Decimal

from money.money import Money
from money.currency import Currency

class MyMoney(Money):

    def __init__(self, amount, currency=Currency.EUR):
        # Allow initialization with an arbitrary number of
        # decimal places but round it to the correct number
        # of decimal places at initialization.
        self._amount = self._round(Decimal(amount), currency)
        self._currency = currency

    def __repr__(self):
        # Omit currency in the repr
        return "{}".format(self._amount)



import pytest

from en16931 import BankInfo


class TestBankInfo:

    def test_initialization_account_bic(self):
        b = BankInfo(account="1234567", bic="AAAABBCCDDD")
        assert b.is_valid()
        assert b.account == "1234567"
        assert b.bic == "AAAABBCCDDD"
    
    def test_initialization_iban(self):
        b = BankInfo(iban="ES661234567321")
        assert b.is_valid()
        assert b.iban == "ES661234567321"

    def test_iban_setter(self):
        b = BankInfo()
        assert not b.is_valid()
        b.iban = "ES661234567321"
        assert b.is_valid()
        assert b.iban == "ES661234567321"

    def test_account_bic_setter(self):
        b = BankInfo()
        assert not b.is_valid()
        b.account = "1234567321"
        b.bic = "AAAABBCCDDD"
        assert b.is_valid()
        assert b.account == "1234567321"
        assert b.bic == "AAAABBCCDDD"

    def test_invalid_bic(self):
        b = BankInfo()
        with pytest.raises(ValueError):
            b.bic = "foo"

    def test_mandate(self):
        b = BankInfo(account="1234567", bic="AAAABBCCDDD",
                     mandate_reference_identifier="123")
        assert b.is_valid()
        assert b.has_mandate_reference()
        assert b.mandate_reference_identifier == "123"
